using GLMakie
GLMakie.activate!()
using dg1d


function map_variablenames(var, output)
  names = last.(split.(output.variables, "."))
  idx = findfirst(v -> v == var, names)
  isnothing(idx) && error("Variable named '$var' found in output")
  return output.variables[idx]
end


function button_labels(font)
  labels = Dict{Symbol,String}()
  if font.family_name == "Unifont" && font.style_name == "Regular"
    labels[:prev] = "\U23F4"
    labels[:next] = "\U23F5"
    labels[:fpsminus] = "-"
    labels[:fpsplus] = "+"
    labels[:playbwrd] = "\U23EA"
    labels[:pause] = "\U23F8"
    labels[:playfwrd] = "\U23E9"
    labels[:stop] = "\U23F9"
    labels[:save] = "\U23FA"
  else
    labels[:prev] = "<"
    labels[:next] = ">"
    labels[:fpsminus] = "-"
    labels[:fpsplus] = "+"
    labels[:playbwrd] = "<="
    labels[:pause] = "||"
    labels[:playfwrd] = "=>"
    labels[:stop] = "stop"
    labels[:save] = "save"
  end
  labels
end


function validator_indexbox(str)
  m1 = match(r"\s*(\d+)(?:/\d+)", str)
end


# This is used in a way to speed up plotting of DG solutions with 
# Makie's lines, scatterlines, etc. methods.
# Compare these two implementations:
#
# ```julia
# x, v = coordinates(...), function_values(...)
# @assert size(x) == size(v)
# Npts, K = size(x)
#
# f = Figure()
# ax = Axis(f[1,1])
# for k = 1:K
#   lines!(ax, x[:,k], v[:,k])
# end
#
# ax = Axis(f[1,2])
# x_w_nan, v_w_nan = nan_separated_storage(size(x)), nan_separated_storage(size(v))
# x_w_nan[1:end-1,:] .= x
# v_w_nan[1:end-1,:] .= v
# lines!(ax, x_w_nan[:], v_w_nan[:])
# ```
#
# The latter performs much faster. Perhaps I should submit a PR to Makie 
# that allows to pass an additional array with indices at which position one should
# splice in NaNs. E.g. above should then be written as
# ```julia
# idx_separate = Npts * collect(1:K)
# ax = Axis(f[1,3])
# lines!(ax, x[:], v[:], idx_separate)
# ```
function nan_separated_storage(m::Mesh1d)
  Npts = m.element.Npts
  K, = m.tree.dims
  arr = zeros(Float64, Npts+1, K)
  arr[end, :] .= NaN
  return arr
end
function nan_separated_storage(m::Mesh2d)
  Npts = m.element.Npts
  K, _ = m.tree.dims
  arr = zeros(Float64, Npts+1, K)
  arr[end, :] .= NaN
  return arr
end


function make_movie(dir;
    fps::Int64=30, ncols::Union{Int,Nothing}=nothing, axis_equal::Bool=false,
    xlimits::Union{Nothing,NTuple{2}}=nothing,
    ylimits::Union{Nothing,NTuple{2}}=nothing)

  dir = abspath(joinpath(".",dir))
  tomlfile = get_parameter_file(dir)
  pars = TOML.parsefile(joinpath(dir, tomlfile))
  project, parameters = dg1d.load_parameters(joinpath(dir, tomlfile))
  mesh_prms = parameters["Mesh"]
  mesh = dg1d.Mesh(N=mesh_prms["n"],
                   K=mesh_prms["k"],
                   CFL=mesh_prms["cfl"],
                   range=mesh_prms["range"],
                   basis=lowercase(mesh_prms["basis"]),
                   periodic=mesh_prms["periodic"])

  # mesh = make_Mesh(trim_headers(parameters, "Mesh"))

  # read data
  output = Output(joinpath(dir, "output.h5"))
  variables = output.variables
  varnames = [ last(split(var, ".")) for var in variables ]
  nvars = length(variables)
  ts = output.times
  Nts = length(ts)
  read_variables = [ t -> read_variable(output, var, t) for var in variables ]

  if isnothing(ncols)
    # as a default determine ncols such that each row has at most max_nrows plots
    max_nrows = 4
    ncols = div(nvars, max_nrows, RoundUp)
  else
    @assert ncols >= 1
  end
  nrows = div(nvars,ncols,RoundUp)
  axis = (nrows, ncols)
  font = to_font("Unifont") # for media player unicode characters
  fontattrs = (;font=font,fontsize=25)
  labels = button_labels(font)

  # prepare axes
  fig = Figure(resolution=(1600,1200))
  t = Observable(ts[1])
  fig[1,1] = headergrid   = GridLayout()
  fig[2,1] = axisgrid     = GridLayout()
  axs = if ncols == 1
    [ Axis(axisgrid[i,1], xlabel="x", ylabel=v) for (i,v) in enumerate(varnames) ]
  else
    axs = Axis[]
    CI = CartesianIndices(axis)
    for i = 1:nvars
      r, c = CI[i].I
      push!(axs, Axis(axisgrid[r,c], xlabel="x", ylabel=varnames[i]))
    end
    axs
  end

  # plot data
  L = layout(mesh)
  Npts,K = L
  plt_xdata = nan_separated_storage(mesh)
  plt_xdata[1:end-1,:] .= mesh.x
  ybuffers  = [ nan_separated_storage(mesh) for _ = 1:nvars ]
  plt_ydata = [ Observable(yb[:]) for yb in ybuffers ]
  for i = 1:nvars
    ybuffers[i][1:end-1,:] .= reshape(read_variables[i](1#=use first data set=#), L)
    plt_ydata[i][] = ybuffers[i][:]
    lines!(axs[i], plt_xdata[:], plt_ydata[i])
    axs[i].limits[] = (xlimits, ylimits)
  end

  ### setup header section

  ## stepper controls
  current_t   = Observable(ts[1])
  slider      = headergrid[1,1:13] = Slider(fig, range=1:Nts, startvalue=1)
  prevbt      = headergrid[2,1]  = Button(fig, label=labels[:prev], font=font, fontsize=35)
  idxbox      = headergrid[2,2]  = Textbox(fig, stored_string="1", validator=Int64,
                                          displayed_string="1", width=Fixed(50),
                                          halign=:center, #reset_on_defocuse=true,
                                          defocus_on_submit=false)
  nextbt      = headergrid[2,3]  = Button(fig, label=labels[:next], font=font, fontsize=35)
  playbwrdbt  = headergrid[2,4]  = Button(fig, label=labels[:playbwrd], font=font, fontsize=35)
  pausebt     = headergrid[2,5]  = Button(fig, label=labels[:pause], font=font, fontsize=35)
  playfwrdbt  = headergrid[2,6]  = Button(fig, label=labels[:playfwrd], font=font, fontsize=35)
  stopbt      = headergrid[2,7]  = Button(fig, label=labels[:stop], font=font, fontsize=35)
  savebt      = headergrid[2,8]  = Button(fig, label=labels[:save], font=font, fontsize=35)
  fpsminusbt  = headergrid[2,9]  = Button(fig, label=labels[:fpsminus], font=font, fontsize=35)
  fpsbox      = headergrid[2,10] = Label(fig, text="FPS", tellwidth=true)
  fpsplusbt   = headergrid[2,11] = Button(fig, label=labels[:fpsplus], font=font, fontsize=35)
  label       = headergrid[2,12] = Label(fig, @lift("t = $($current_t)"), tellwidth=true,
                                        halign=:left)
  headergrid[2,13] = GLMakie.Box(fig, visible=false)

  function step_frame(idx)
    ti = ts[idx]
    current_t[] = ti
    for n = 1:nvars
      ybuffers[n][1:end-1,:] .= reshape(read_variables[n](idx), L)
      plt_ydata[n][] = ybuffers[n][:]
      axs[n].limits[] = (xlimits, ylimits)
    end
  end

  on(slider.value) do i
    idxbox.displayed_string[] = string(i)
    step_frame(i)
  end

  on(prevbt.clicks) do n
    value = slider.value[]
    if value > 1
      set_close_to!(slider, value-1)
    end
  end

  on(nextbt.clicks) do n
    value = slider.value[]
    if value < Nts
      set_close_to!(slider, value+1)
    end
  end

  on(idxbox.stored_string) do s
    idx = parse(Int64, s)
    idx = max(idx, 1)
    idx = min(idx, Nts)
    set_close_to!(slider, idx)
  end

  # # FPS controls
  fps = Observable{Int64}(fps)
  on(fpsminusbt.clicks) do n
    fps[] = max(fps[]-10, 10)
  end
  on(fpsplusbt.clicks) do n
    fps[] = min(fps[]+10, 100)
  end

  tasks = Dict{Symbol,Task}()

  interrupt = false
  function play_frames_forward()
    while slider.value[] < Nts
      interrupt && return
      notify(nextbt.clicks)
      sleep(1/fps[])
    end
  end
  tasks[:play_forward] = Task(play_frames_forward)

  function play_frames_backward()
    while slider.value[] > 1
      interrupt && return
      notify(prevbt.clicks)
      sleep(1/fps[])
    end
  end
  tasks[:play_backward] = Task(play_frames_backward)

  function record_frames()
    tmp_value = slider.value[]
    moviename = joinpath(dir, "movie.mp4")
    record(fig, moviename, framerate=fps[]) do io
      @warn "Recording frames, don't resize window or touch controls ..."
      for i = 1:Nts
        set_close_to!(slider, i)
        recordframe!(io)
      end
      @info "Saving movie to '$moviename' ..."
    end
    @info "Done recording"
    # restore initial state
    set_close_to!(slider, tmp_value)
  end
  tasks[:record] = Task(record_frames)

  isrunning(t) = istaskstarted(t) && !istaskdone(t) && !istaskfailed(t)

  on(playfwrdbt.clicks) do n
    if !isrunning(tasks[:record])
      isrunning(tasks[:play_backward]) && notify(pausebt.clicks)
      interrupt = false
      tasks[:play_forward] = Task(play_frames_forward)
      schedule(tasks[:play_forward])
    end
  end

  on(playbwrdbt.clicks) do n
    if !isrunning(tasks[:record])
      isrunning(tasks[:play_forward]) && notify(pausebt.clicks)
      interrupt = false
      tasks[:play_backward] = Task(play_frames_backward)
      schedule(tasks[:play_backward])
    end
  end

  on(pausebt.clicks) do n
    if any(isrunning.(values(tasks)))
      interrupt = true
      for t in values(tasks)
        isrunning(t) && wait(t)
      end
    end
  end

  on(savebt.clicks) do n
    if istaskdone(tasks[:record])
      tasks[:record] = Task(record_frames) # renew task
    end
    if !istaskstarted(tasks[:record])
      interrupt = false
      notify(pausebt.clicks) # stop any playing tasks
      schedule(tasks[:record])
    end
  end

  return fig
end
